@@ -1,19 +1,25 @@
 use std::env::ArgsOs;
+use std::ffi::CString;
 use std::io::{self, Stderr, Stdout, Write};
+use std::os::raw::c_char;
 use std::process::exit;
+use std::ptr;
 use std::time::Duration;
 
 use crossterm::{
     cursor,
     event::{self, Event, KeyCode, KeyEvent, KeyModifiers},
-    execute, queue, style,
+    execute, queue,
+    style::{self, Attribute},
     terminal::{self, ClearType},
 };
 
 use crate::args::{Parser, HELP};
 use crate::error::Result;
 use crate::finder::Finder;
-use crate::{Error, MatchedPath};
+use crate::matched_path::MatchedPath;
+use crate::starting_point::StartingPoint;
+use crate::Error;
 
 pub fn safe_exit(code: i32, stdout: Stdout, stderr: Stderr) {
     let _ = stdout.lock().flush();
@@ -21,29 +27,39 @@ pub fn safe_exit(code: i32, stdout: Stdout, stderr: Stderr) {
     exit(code)
 }
 
-pub fn entrypoint(args: ArgsOs, stdout: &mut impl Write, stderr: &mut impl Write) -> Result<()> {
+pub fn entrypoint(args: ArgsOs, stdout: &mut impl Write) -> Result<()> {
     let args = Parser::new(args).parse()?;
     if args.help {
-        print_help(stdout)?;
+        print_and_flush(stdout, HELP)?;
+        return Ok(());
+    }
+    if args.version {
+        print_and_flush(
+            stdout,
+            &format!("{} {}\n", env!("CARGO_PKG_NAME"), env!("CARGO_PKG_VERSION")),
+        )?;
         return Ok(());
     }
 
-    let (_, mut rows) = terminal::size()?;
+    let (mut columns, mut rows) = terminal::size()?;
+    let starting_point = StartingPoint::new(args.starting_point)?;
     let mut query = args.query;
-    let mut paths = find_paths(&args.starting_point, &query, rows - 1)?;
+    let mut paths = find_paths(&starting_point, &query, paths_rows(rows))?;
     let mut selection: u16 = 0;
     let mut state = State::QueryChanged;
     initialize_terminal(stdout)?;
 
     loop {
         match state {
-            State::QueryChanged => output_on_terminal(stdout, &query, &paths[..], selection)?,
+            State::QueryChanged => {
+                output_on_terminal(stdout, &query, &paths[..], selection, columns)?
+            }
             State::PathsChanged => {
-                output_on_terminal(stdout, &query, &paths[..], selection)?;
+                output_on_terminal(stdout, &query, &paths[..], selection, columns)?;
                 state = State::Ready;
             }
             State::SelectionChanged => {
-                output_on_terminal(stdout, &query, &paths[..], selection)?;
+                output_on_terminal(stdout, &query, &paths[..], selection, columns)?;
                 state = State::Ready;
             }
             _ => (),
@@ -51,10 +67,19 @@ pub fn entrypoint(args: ArgsOs, stdout: &mut impl Write, stderr: &mut impl Write
 
         if event::poll(Duration::from_millis(300))? {
             let ev = event::read()?;
-            // TODO: Support uppercase
-            if let Event::Key(KeyEvent {
+            if ev
+                == Event::Key(KeyEvent {
+                    code: KeyCode::Char('c'),
+                    modifiers: KeyModifiers::CONTROL,
+                })
+                || ev == Event::Key(KeyCode::Esc.into())
+            {
+                // CTRL-C does not send SIGINT even on UNIX/Linux because it's in Raw mode.
+                // Also, we handle Esc as the same.
+                break;
+            } else if let Event::Key(KeyEvent {
                 code: KeyCode::Char(c),
-                modifiers: KeyModifiers::NONE,
+                modifiers: _,
             }) = ev
             {
                 query.push(c);
@@ -68,7 +93,7 @@ pub fn entrypoint(args: ArgsOs, stdout: &mut impl Write, stderr: &mut impl Write
                 }
                 state = State::SelectionChanged;
             } else if ev == Event::Key(KeyCode::Down.into()) {
-                if selection < (rows - 1) {
+                if selection < paths_rows(rows) - 1 {
                     selection += 1;
                 }
                 state = State::SelectionChanged;
@@ -76,22 +101,13 @@ pub fn entrypoint(args: ArgsOs, stdout: &mut impl Write, stderr: &mut impl Write
                 let path: &MatchedPath = paths.get(selection as usize).unwrap(); // TODO: Do not use unwrap()
                 state = State::Invoke(&path);
                 break;
-            } else if ev
-                == Event::Key(KeyEvent {
-                    code: KeyCode::Char('c'),
-                    modifiers: KeyModifiers::CONTROL,
-                })
-                || ev == Event::Key(KeyCode::Esc.into())
-            {
-                // CTRL-C does not send SIGINT even on UNIX/Linux because it's in Raw mode.
-                // Also, we handle Esc as the same.
-                break;
-            } else if let Event::Resize(_, r) = ev {
+            } else if let Event::Resize(c, r) = ev {
+                columns = c;
                 rows = r;
                 state = State::PathsChanged;
             }
         } else if let State::QueryChanged = state {
-            paths = find_paths(&args.starting_point, &query, rows - 1)?;
+            paths = find_paths(&starting_point, &query, paths_rows(rows))?;
             state = State::PathsChanged;
             selection = 0;
         }
@@ -101,17 +117,7 @@ pub fn entrypoint(args: ArgsOs, stdout: &mut impl Write, stderr: &mut impl Write
     terminal::disable_raw_mode()?;
 
     if let State::Invoke(path) = state {
-        // TODO: Decide which we should pass: relative or absolute.
-        let path = &path.relative;
-        let output = invoke(&args.exec, &path)?;
-        stdout.write_all(&output.stdout)?;
-        stderr.write_all(&output.stderr)?;
-        if !output.status.success() {
-            return Err(Error::exec(&format!(
-                "Failed to execute command `{} {}`",
-                args.exec, path,
-            )));
-        }
+        invoke(&args.exec, path.absolute())?;
     }
     Ok(())
 }
@@ -124,8 +130,13 @@ enum State<'a> {
     Invoke(&'a MatchedPath),
 }
 
-fn print_help(buffer: &mut impl Write) -> io::Result<()> {
-    buffer.write_all(format!("{}\n", HELP).as_bytes())?;
+fn paths_rows(row: u16) -> u16 {
+    // TODO: raise an error when the number of rows is too small.
+    row - 1
+}
+
+fn print_and_flush(buffer: &mut impl Write, content: &str) -> io::Result<()> {
+    buffer.write_all(content.as_bytes())?;
     buffer.flush()
 }
 
@@ -141,6 +152,7 @@ fn output_on_terminal(
     query: &str,
     paths: &[MatchedPath],
     selection: u16,
+    max_columns: u16,
 ) -> Result<()> {
     queue!(
         stdout,
@@ -154,38 +166,58 @@ fn output_on_terminal(
     for (idx, path) in paths.iter().enumerate() {
         let idx = idx as u16;
         let prefix = if idx == selection { "> " } else { "  " };
-        queue!(
-            stdout,
-            style::Print(format!("{}{}", prefix, path)),
-            cursor::MoveToNextLine(1),
-        )?;
+        queue!(stdout, style::Print(prefix))?;
+
+        let max_columns: usize = max_columns as usize - 2; // The prefix "> " requires 2 columns.
+        for chunk in path.chunks(max_columns) {
+            if chunk.matched() {
+                queue!(
+                    stdout,
+                    style::SetAttribute(Attribute::Bold),
+                    style::Print(format!("{}", chunk)),
+                    style::SetAttribute(Attribute::Reset),
+                )?;
+            } else {
+                queue!(stdout, style::Print(format!("{}", chunk)))?;
+            }
+        }
+        queue!(stdout, cursor::MoveToNextLine(1))?;
     }
     queue!(stdout, cursor::RestorePosition)?;
     stdout.flush()?;
     Ok(())
 }
 
-fn find_paths(starting_point: &str, query: &str, limit: u16) -> Result<Vec<MatchedPath>> {
+fn find_paths(starting_point: &StartingPoint, query: &str, limit: u16) -> Result<Vec<MatchedPath>> {
     let mut paths = Vec::with_capacity(100); // TODO: Tune this capacity later.
 
-    // TODO: Sort
-    for path in Finder::new(starting_point, query)?.take(limit.into()) {
+    for path in Finder::new(starting_point, query)? {
+        // TODO: Shouldn't we stop iteration when some paths returns Err?
         let path = path?;
         paths.push(path);
     }
-    Ok(paths)
+
+    paths.sort();
+    Ok(paths.into_iter().take(limit.into()).collect())
 }
 
-fn invoke(exec: &str, path: &str) -> Result<std::process::Output> {
-    let mut exec = exec.split_whitespace();
-    let program = exec
-        .next()
-        .ok_or_else(|| Error::args("\"exec\" cannot be processed with empty string"))?;
-    let mut command = std::process::Command::new(program);
-    for arg in exec {
-        command.arg(arg);
+/// Invoke the specified command with the selected path.
+fn invoke(exec: &str, path: &str) -> Result<()> {
+    let mut cstrings: Vec<CString> = Vec::with_capacity(10); // TODO: Why is it 10?
+    for arg in exec.split_whitespace() {
+        cstrings.push(CString::new(arg)?);
     }
-    command.arg(path);
-    let output = command.output()?;
-    Ok(output)
+    cstrings.push(CString::new(path)?);
+    let argv: Vec<*const c_char> = cstrings
+        .iter()
+        .map(|c| c.as_ptr())
+        .chain(std::iter::once(ptr::null()))
+        .collect();
+
+    let errno = unsafe { libc::execvp(cstrings[0].as_ptr(), argv.as_ptr()) };
+
+    Err(Error::exec(&format!(
+        "`{} {}` failed and returned {}",
+        exec, path, errno
+    )))
 }
